/*
 * Copyright (C) 1997-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file client_common.c Common functions of para_client and para_audiod. */

#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "error.h"
#include "list.h"
#include "sched.h"
#include "client.cmdline.h"
#include "crypt.h"
#include "rc4.h"
#include "net.h"
#include "fd.h"
#include "string.h"
#include "client.cmdline.h"
#include "client.h"

/*
 * Rc4-encrypt data before sending.
 *
 * \param len The number of bytes to encrypt.
 * \param indata Pointer to the input data of length \a len to be encrypted.
 * \param outdata Result-pointer that holds the encrypted data.
 * \param private_data Contains the rc4 key.
 */
static void rc4_send(unsigned long len, const unsigned char *indata,
		unsigned char *outdata, void *private_data)
{
	struct client_task *ct = private_data;
	RC4(&ct->rc4_send_key, len, indata, outdata);
}

/*
 * Rc4-decrypt received data.
 *
 * Parameters are identical to those of rc4_send.
 */
static void rc4_recv(unsigned long len, const unsigned char *indata,
		unsigned char *outdata, void *private_data)
{
	struct client_task *ct = private_data;
	RC4(&ct->rc4_recv_key, len, indata, outdata);
}

/**
 * Close the connection to para_server and free all resources.
 *
 * \param ct Pointer to the client data.
 *
 * \sa client_open.
 */
void client_close(struct client_task *ct)
{
	if (!ct)
		return;
	if (ct->fd >= 0) {
		disable_crypt(ct->fd);
		close(ct->fd);
	}
	free(ct->user);
	free(ct->config_file);
	free(ct->key_file);
	client_cmdline_parser_free(&ct->conf);
	free(ct);
}

/**
 * The preselect hook for server commands.
 *
 * \param s Pointer to the scheduler.
 * \param t Pointer to the task struct for this command.
 *
 * The task pointer must contain a pointer to the initialized client data
 * structure as it is returned by client_open().
 *
 * This function checks the state of the connection and adds the file descriptor
 * of the connection to the read or write fd set of \a s accordingly.
 *
 * \sa register_task() client_open(), struct sched, struct task.
 */
static void client_pre_select(struct sched *s, struct task *t)
{
	struct client_task *ct = container_of(t, struct client_task, task);

	ct->check_r = 0;
	ct->check_w = 0;
	if (ct->fd < 0)
		return;
	switch (ct->status) {
	case CL_CONNECTED:
	case CL_SENT_AUTH:
	case CL_SENT_CH_RESPONSE:
	case CL_SENT_COMMAND:
		para_fd_set(ct->fd, &s->rfds, &s->max_fileno);
		ct->check_r = 1;
		return;

	case CL_RECEIVED_WELCOME:
	case CL_RECEIVED_CHALLENGE:
	case CL_RECEIVED_PROCEED:
		para_fd_set(ct->fd, &s->wfds, &s->max_fileno);
		ct->check_w = 1;
		return;

	case CL_RECEIVING:
		if (ct->loaded < CLIENT_BUFSIZE - 1) {
			para_fd_set(ct->fd, &s->rfds, &s->max_fileno);
			ct->check_r = 1;
		}
		return;
	case CL_SENDING:
		if (!ct->in_loaded) /* stdin task not yet started */
			return;
		if (*ct->in_loaded) {
			PARA_INFO_LOG("loaded: %zd\n", *ct->in_loaded);
			para_fd_set(ct->fd, &s->wfds, &s->max_fileno);
			ct->check_w = 1;
		} else {
			if (*ct->in_error) {
				t->error = *ct->in_error;
				s->timeout.tv_sec = 0;
				s->timeout.tv_usec = 1;
			}
		}
		return;
	}
}

static ssize_t client_recv_buffer(struct client_task *ct)
{
	ssize_t ret = recv_buffer(ct->fd, ct->buf + ct->loaded,
		CLIENT_BUFSIZE - ct->loaded);
	if (!ret)
		return -E_SERVER_EOF;
	if (ret > 0)
		ct->loaded += ret;
	return ret;

}

/**
 * The post select hook for client commands.
 *
 * \param s Pointer to the scheduler.
 * \param t Pointer to the task struct for this command.
 *
 * Depending on the current state of the connection and the status of the read
 * and write fd sets of \a s, this function performs the necessary steps to
 * authenticate the connection, to send the command given by \a t->private_data
 * and to receive para_server's output, if any.
 *
 * \sa struct sched, struct task.
 */
static void client_post_select(struct sched *s, struct task *t)
{
	struct client_task *ct = container_of(t, struct client_task, task);

	t->error = 0;
	if (ct->fd < 0)
		return;
	if (!ct->check_r && !ct->check_w)
		return;
	if (ct->check_r && !FD_ISSET(ct->fd, &s->rfds))
		return;
	if (ct->check_w && !FD_ISSET(ct->fd, &s->wfds))
		return;
	switch (ct->status) {
	case CL_CONNECTED: /* receive welcome message */
		t->error = client_recv_buffer(ct);
		if (t->error > 0)
			ct->status = CL_RECEIVED_WELCOME;
		return;
	case CL_RECEIVED_WELCOME: /* send auth command */
		sprintf(ct->buf, "auth %s%s", ct->conf.plain_given?
			"" : "rc4 ", ct->user);
		PARA_INFO_LOG("--> %s\n", ct->buf);
		t->error = send_buffer(ct->fd, ct->buf);
		if (t->error >= 0)
			ct->status = CL_SENT_AUTH;
		return;
	case CL_SENT_AUTH: /* receive challenge number */
		ct->loaded = 0;
		t->error = client_recv_buffer(ct);
		if (t->error < 0)
			return;
		if (t->error != 64) {
			t->error = -E_INVALID_CHALLENGE;
			PARA_ERROR_LOG("received the following: %s\n", ct->buf);
			return;
		}
		PARA_INFO_LOG("<-- [challenge]\n");
		/* decrypt challenge number */
		t->error = para_decrypt_challenge(ct->key_file, &ct->challenge_nr,
			(unsigned char *) ct->buf, 64);
		if (t->error > 0)
			ct->status = CL_RECEIVED_CHALLENGE;
		return;
	case CL_RECEIVED_CHALLENGE: /* send decrypted challenge */
		PARA_INFO_LOG("--> %lu\n", ct->challenge_nr);
		t->error = send_va_buffer(ct->fd, "%s%lu", CHALLENGE_RESPONSE_MSG,
			ct->challenge_nr);
		if (t->error > 0)
			ct->status = CL_SENT_CH_RESPONSE;
		return;
	case CL_SENT_CH_RESPONSE: /* read server response */
		{
		size_t bytes_received;
		unsigned char rc4_buf[2 * RC4_KEY_LEN] = "";
		ct->loaded = 0;
		t->error = client_recv_buffer(ct);
		if (t->error < 0)
			return;
		bytes_received = t->error;
		PARA_DEBUG_LOG("++++ server info ++++\n%s\n++++ end of server "
			"info ++++\n", ct->buf);
		/* check if server has sent "Proceed" message */
		t->error = -E_CLIENT_AUTH;
		if (!strstr(ct->buf, PROCEED_MSG))
			return;
		t->error = 0;
		ct->status = CL_RECEIVED_PROCEED;
		if (bytes_received < PROCEED_MSG_LEN + 32)
			return;
		PARA_INFO_LOG("decrypting session key\n");
		t->error = para_decrypt_buffer(ct->key_file, rc4_buf,
			(unsigned char *)ct->buf + PROCEED_MSG_LEN + 1,
			bytes_received - PROCEED_MSG_LEN - 1);
		if (t->error < 0)
			return;
		RC4_set_key(&ct->rc4_send_key, RC4_KEY_LEN, rc4_buf);
		RC4_set_key(&ct->rc4_recv_key, RC4_KEY_LEN, rc4_buf + RC4_KEY_LEN);
		enable_crypt(ct->fd, rc4_recv, rc4_send, ct);
		}
	case CL_RECEIVED_PROCEED: /* concat args and send command */
		{
		int i;
		char *command = NULL;
		for (i = 0; i < ct->conf.inputs_num; i++) {
			char *tmp = command;
			command = make_message("%s\n%s", command?
				command : "", ct->conf.inputs[i]);
			free(tmp);
		}
		command = para_strcat(command, EOC_MSG "\n");
		PARA_DEBUG_LOG("--> %s\n", command);
		t->error = send_buffer(ct->fd, command);
		free(command);
		if (t->error > 0)
			ct->status = CL_SENT_COMMAND;
		return;
		}
	case CL_SENT_COMMAND:
		ct->loaded = 0;
		t->error = client_recv_buffer(ct);
		if (t->error < 0)
			return;
		if (strstr(ct->buf, AWAITING_DATA_MSG))
			ct->status = CL_SENDING;
		else
			ct->status = CL_RECEIVING;
		return;
	case CL_SENDING: /* FIXME: might block */
		PARA_INFO_LOG("loaded: %zd\n", *ct->in_loaded);
		t->error = send_bin_buffer(ct->fd, ct->inbuf, *ct->in_loaded);
		if (t->error < 0)
			return;
		*ct->in_loaded = 0;
		return;
	case CL_RECEIVING:
		t->error = client_recv_buffer(ct);
		return;
	}
}

/* connect to para_server and register the client task */
static int client_connect(struct client_task *ct)
{
	int ret;

	ct->fd = -1;
	ret = makesock(AF_UNSPEC, IPPROTO_TCP, 0, ct->conf.hostname_arg,
		ct->conf.server_port_arg);
	if (ret < 0)
		return ret;
	ct->fd = ret;
	ct->status = CL_CONNECTED;
	ret = mark_fd_nonblocking(ct->fd);
	if (ret < 0)
		goto err_out;
	ct->task.pre_select = client_pre_select;
	ct->task.post_select = client_post_select;
	sprintf(ct->task.status, "client");
	register_task(&ct->task);
	return 1;
err_out:
	close(ct->fd);
	ct->fd = -1;
	return ret;
}

/**
 * Open connection to para_server.
 *
 * \param argc Usual argument count.
 * \param argv Usual argument vector.
 * \param ct_ptr Points to dynamically allocated and initialized client task
 * struct upon successful return.
 *
 * Check the command line options given by \a argc and argv, set default values
 * for user name and rsa key file, read further option from the config file.
 * Finally, establish a connection to para_server.
 *
 * \return Standard.
 */
int client_open(int argc, char *argv[], struct client_task **ct_ptr)
{
	char *home = para_homedir();
	struct stat statbuf;
	int ret;
	struct client_task *ct = para_calloc(sizeof(struct client_task));

	*ct_ptr = ct;
	ct->fd = -1;
	ret = client_cmdline_parser(argc, argv, &ct->conf);
	HANDLE_VERSION_FLAG("client", ct->conf);
	ret = -E_CLIENT_SYNTAX;
	if (!ct->conf.inputs_num)
		goto out;
	ct->user = ct->conf.user_given?
		para_strdup(ct->conf.user_arg) : para_logname();

	ct->key_file = ct->conf.key_file_given?
		para_strdup(ct->conf.key_file_arg) :
		make_message("%s/.paraslash/key.%s", home, ct->user);

	ct->config_file = ct->conf.config_file_given?
		para_strdup(ct->conf.config_file_arg) :
		make_message("%s/.paraslash/client.conf", home);
	ret = stat(ct->config_file, &statbuf);
	if (ret && ct->conf.config_file_given) {
		ret = -E_NO_CONFIG;
		goto out;
	}
	if (!ret) {
		struct client_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0
		};
		client_cmdline_parser_config_file(ct->config_file,
			&ct->conf, &params);
	}
	ret = 1;
	PARA_INFO_LOG("loglevel: %d\n", ct->conf.loglevel_arg);
	PARA_INFO_LOG("config_file: %s\n", ct->config_file);
	PARA_INFO_LOG("key_file: %s\n", ct->key_file);
	PARA_NOTICE_LOG("connecting %s:%d\n", ct->conf.hostname_arg,
		ct->conf.server_port_arg);
	ret = client_connect(ct);
out:
	free(home);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		client_close(ct);
		*ct_ptr = NULL;
	}
	return ret;
}

