/*
 * Copyright (C) 2006-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file dccp_recv.c paraslash's dccp receiver */

/*
 * based on client.c of dccp-cs-0.01.tar.bz2,
 * (C) 2005 Ian McDonald <imcdnzl@gmail.com>
 */

#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "error.h"
#include "list.h"
#include "sched.h"
#include "recv.h"
#include "string.h"
#include "net.h"
#include "fd.h"

#include "dccp_recv.cmdline.h"

/** the size of the output buffer */
#define DCCP_BUFSIZE 40960

/**
 * data specific to the dccp receiver
 *
 * \sa receiver receiver_node
 */
struct private_dccp_recv_data {
	/** the file descriptor for the dccp socket */
	int fd;
};


static void dccp_recv_close(struct receiver_node *rn)
{

	struct private_dccp_recv_data *pdd = rn->private_data;

	if (pdd && pdd->fd > 0)
		close(pdd->fd);
	free(rn->buf);
	rn->buf = NULL;
	free(rn->private_data);
	rn->private_data = NULL;
}


static int dccp_recv_open(struct receiver_node *rn)
{
	struct private_dccp_recv_data *pdd;
	struct dccp_recv_args_info *conf = rn->conf;
	int fd, ret = makesock(AF_UNSPEC, IPPROTO_DCCP, 0, conf->host_arg,
		conf->port_arg);

	if (ret < 0)
		return ret;
	fd = ret;
	/*
	 * Disable unused CCIDs: the receiver does not send any application
	 * data to the server. By shutting down this unused path we reduce
	 * internal processing costs, as the unused CCIDs (in the kernel) are
	 * then bypassed.
	 */
	if (shutdown(fd, SHUT_WR) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	ret = mark_fd_nonblocking(fd);
	if (ret < 0)
		goto err;
	rn->buf = para_calloc(DCCP_BUFSIZE);
	rn->private_data = pdd = para_calloc(sizeof(struct private_dccp_recv_data));
	pdd->fd = fd;
	return 1;
err:
	close(fd);
	return ret;
}

static void dccp_shutdown(void)
{
	; /* nothing to do */
}

static void *dccp_recv_parse_config(int argc, char **argv)
{
	struct dccp_recv_args_info *tmp = para_calloc(sizeof(struct dccp_recv_args_info));

	if (!dccp_recv_cmdline_parser(argc, argv, tmp))
		return tmp;
	free(tmp);
	return NULL;
}

static void dccp_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_dccp_recv_data *pdd = rn->private_data;

	t->error = 0;
	para_fd_set(pdd->fd, &s->rfds, &s->max_fileno);
}

static void dccp_recv_post_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_dccp_recv_data *pdd = rn->private_data;

	if (rn->output_error && *rn->output_error < 0) {
		t->error = *rn->output_error;
		return;
	}
	if (!FD_ISSET(pdd->fd, &s->rfds))
		return; /* nothing to do */
	if (rn->loaded >= DCCP_BUFSIZE) {
		t->error = -E_DCCP_OVERRUN;
		return;
	}
	t->error = recv_bin_buffer(pdd->fd, rn->buf + rn->loaded,
		DCCP_BUFSIZE - rn->loaded);
	if (t->error > 0) {
		rn->loaded += t->error;
		return;
	}
	if (!t->error)
		t->error = -E_RECV_EOF;
}

/**
 * The init function of the dccp receiver.
 *
 * \param r Pointer to the receiver struct to initialize.
 *
 * Initialize all function pointers of \a r.
 */
void dccp_recv_init(struct receiver *r)
{
	r->shutdown = dccp_shutdown;
	r->open = dccp_recv_open;
	r->close = dccp_recv_close;
	r->pre_select = dccp_recv_pre_select;
	r->post_select = dccp_recv_post_select;
	r->parse_config = dccp_recv_parse_config;
}
